// -*- C++ -*-

// (c) COPYRIGHT URI/MIT 1999
// Please read the full copyright statement in the file COPYRIGHT.
//
// Authors:
//      jhrg,jimg       James Gallagher (jgallagher@gso.uri.edu)

#include "config_jg.h"

#ifdef __GNUG__
#pragma implementation
#endif

#include <stdlib.h>

#include "Error.h"
#include "JGFloat32.h"
#include "dods-limits.h"
#include "jgofs.h"

Float32 *
NewFloat32(const string &n)
{
    return new JGFloat32(n);
}

JGFloat32::JGFloat32(const string &n) : Float32(n)
{
    _missing_val = DODS_FLT_MAX;
}

BaseType *
JGFloat32::ptr_duplicate()
{
    return new JGFloat32(*this); // Copy ctor calls duplicate to do the work
}

// The following mfunc is untested. jhrg 12/31/97.
 
bool
JGFloat32::read(const string &dataset)
{
  for (int j = 0; j < nVars; ++j) {
    if (name() == varName[j]) {
      char *errptr;
      dods_float32 df = strtod(varValue[j], &errptr);

      if ( errptr != (char *)&(varValue[j]) )
	val2buf(&df);
      else 
	val2buf(&_missing_val);	      

      set_read_p(true);
      return true;
    }
  }

  // If we're here, the variable was not found. throw and error.
  string s = (string)"Variable " + name() 
    + " wasn't located in JGOFS object " + dataset + ".";
  throw Error(unknown_error, s);
}

dods_float32
JGFloat32::get_missing_value() const
{
    return _missing_val;
}

void
JGFloat32::set_missing_value(const dods_float32 &v)
{
    _missing_val = v;
}

void
JGFloat32::set_missing_value(const  string &attribute)
{
    _missing_val = static_cast<float>(strtod(attribute.c_str(), (char **)0));
}

// $Log: JGFloat32.cc,v $
// Revision 1.6  2003/01/16 07:57:34  jimg
// Merged changes from release-3-2-6.
// Removed a number of files used for the client library ans sample clients.
//
// Revision 1.4.4.5  2003/01/16 01:58:57  jimg
// Changed set_missing_value(const string&) so that it uses strtod and casts the
// result to a float. The function strtof() is not present on many machines,
// even Linux (although it's listed in the man pages...).
//
// Revision 1.4.4.4  2003/01/14 05:36:46  jimg
// I added a set_missing_value(const string &) method. This lets the code
// in jgdds.cc that builds a DDS (which may contain types other than string
// if that has been configured in the dataset's DAS) set up the missing
// values using information from the DAS without doing string to value
// conversions for each numeric type in a big switch statement.
//
// Revision 1.4.4.3  2002/01/15 21:05:08  jimg
// I switched from the numeric constants in values.h/limits.h to those in
// dods-limits.h.
//
// Revision 1.5  2001/08/27 17:41:03  jimg
// Merged with release-3-2-2
//
// Revision 1.4.4.2  2001/08/21 15:26:50  dan
// Updated names of the set/get accessor methods for modifying the missing_value
// member field, to make the naming convention consistent with the current
// DAP usage.
//
// Revision 1.4.4.1  2001/05/31 19:12:56  dan
// Modified return value from read method, now returns true.
//
// Revision 1.4  2000/10/10 23:36:14  jimg
// Added exception processing code.
// Added throws to the read methods.
// Changed the definition of read to match the dap library.
// Moved CVS Log entries to the end of files.
//
// Revision 1.3  1999/11/05 01:17:23  jimg
// Result of merge with release-3-1-0
//
// Revision 1.2.6.2  1999/10/30 19:44:17  dan
// Added explicit (char *) cast in error-checking within read().
//
// Revision 1.2.6.1  1999/10/24 18:31:26  dan
// Modified to support type conversion from JGOFS
// string variables to dods_float32.
//
// Revision 1.2  1999/05/05 22:57:18  jimg
// Merged Brent's String --> string changes with Dan's changes
//
// Revision 1.1  1999/03/27 01:00:34  jimg
// Added
//
